use std::fmt::Debug;

use super::diff::DiffStatus;
use super::sexp::Token;
use super::util::get_property_quoted_string;

#[derive(Debug)]
pub(crate) enum EmacsField<'s> {
    Required(&'s str),
    #[allow(dead_code)]
    Optional(&'s str),
}

/// Do no comparison.
///
/// This is for when you want to acknowledge that a field exists in the emacs token, but you do not have any validation for it when using the compare_properties!() macro. Ideally, this should be kept to a minimum since this represents untested values.
#[allow(dead_code)]
pub(crate) fn compare_noop<'b, 's, 'x, R, RG>(
    _emacs: &'b Token<'s>,
    _rust_node: R,
    _emacs_field: &'x str,
    _rust_value_getter: RG,
) -> Result<Option<(DiffStatus, Option<String>)>, Box<dyn std::error::Error>> {
    Ok(None)
}

/// Do no comparison.
///
/// This is for when you want to acknowledge that a field exists in the emacs token, but you do not have any validation for it when using the compare_properties!() macro. Ideally, this should be kept to a minimum since this represents untested values.
#[allow(dead_code)]
pub(crate) fn compare_identity() -> () {
    ()
}

pub(crate) fn compare_property_quoted_string<'b, 's, 'x, R, RG: Fn(R) -> Option<&'s str>>(
    emacs: &'b Token<'s>,
    rust_node: R,
    emacs_field: &'x str,
    rust_value_getter: RG,
) -> Result<Option<(DiffStatus, Option<String>)>, Box<dyn std::error::Error>> {
    let value = get_property_quoted_string(emacs, emacs_field)?;
    let rust_value = rust_value_getter(rust_node);
    if !rust_value.eq(&value.as_ref().map(String::as_str)) {
        let this_status = DiffStatus::Bad;
        let message = Some(format!(
            "{} mismatch (emacs != rust) {:?} != {:?}",
            emacs_field, value, rust_value
        ));
        Ok(Some((this_status, message)))
    } else {
        Ok(None)
    }
}
